#include "triangle.h"
#include <boost/python.hpp>
#include <vector>
#include <stdexcept>
#include <iostream>




using namespace boost::python;
using namespace std;




class tSizeChangeNotifier;




class tSizeChangeNotificationReceiver
{
  public:
    virtual ~tSizeChangeNotificationReceiver()
    { }
    virtual void notifySizeChange(tSizeChangeNotifier *master, unsigned size) = 0;
};




class tSizeChangeNotifier
{
    typedef vector<tSizeChangeNotificationReceiver *> tNotificationReceiverList;
    tNotificationReceiverList NotificationReceivers;

  public:
    virtual ~tSizeChangeNotifier()
    { }
    virtual unsigned size() const = 0;
    virtual void setSize(unsigned size)
    {
      tNotificationReceiverList::iterator first = NotificationReceivers.begin(),
      last = NotificationReceivers.end();
      while (first != last)
	(*first++)->notifySizeChange(this, size);
    }

    void registerForNotification(tSizeChangeNotificationReceiver *rec)
    {
      NotificationReceivers.push_back(rec);
    }

    void unregisterForNotification(tSizeChangeNotificationReceiver *rec)
    {
      tNotificationReceiverList::iterator first = NotificationReceivers.begin(),
      last = NotificationReceivers.end();
      while (first != last)
      {
	if (rec == *first)
	{
	  NotificationReceivers.erase(first);
	  return;
	}
	first++;
      }
    }
};




class tVertex : public boost::noncopyable
{
  public:
    REAL	*Data;

  public:
    tVertex(REAL *data)
    : Data(data)
    {
    }

    REAL x() { return Data[ 0 ]; }
    REAL y() { return Data[ 1 ]; }
};




template<class ElementT> 
class tForeignArray : public tSizeChangeNotifier, public tSizeChangeNotificationReceiver,
  public boost::noncopyable
{
    ElementT	*&Contents;
    int		&NumberOf;
    unsigned	Unit;
    tSizeChangeNotifier *SlaveTo;
    string      Name;

  public:
    tForeignArray(const string &name, 
        ElementT *&cts, int &number_of, unsigned unit = 1, tSizeChangeNotifier *slave_to = NULL)
      : Contents(cts), NumberOf(number_of), Unit(unit), SlaveTo(slave_to), Name(name)
    {
      Contents = NULL;
      if (SlaveTo)
      {
	SlaveTo->registerForNotification(this);
	setSizeInternal(SlaveTo->size());
      }
      else
	setSize(0);
    }

    ~tForeignArray()
    {
      if (SlaveTo)
	SlaveTo->unregisterForNotification(this);

      if (Contents)
	free(Contents);
      Contents = NULL;

      if (!SlaveTo)
	NumberOf = 0;
    }

    unsigned size() const
    {
      return NumberOf;
    }

    unsigned unit() const
    {
      return Unit;
    }

    void deallocate()
    {
      if (Contents != NULL)
	free(Contents);
      Contents = NULL;
    }

    void setSize(unsigned size)
    {
      if (SlaveTo)
	throw runtime_error("sizes of slave arrays cannot be changed");
      else
	setSizeInternal(size);
    }
    
    void setup()
    {
      if (!SlaveTo)
	throw runtime_error("cannot setup non-slave array");
      else
	setSizeInternal(NumberOf);

    }

    void notifySizeChange(tSizeChangeNotifier *master, unsigned size)
    {
      setSizeInternal(size);
    }

    void setSizeInternal(unsigned size)
    {
      if (!SlaveTo)
	NumberOf = size;
      
      if (Contents != NULL)
	free(Contents);

      if (size == 0 || Unit == 0)
	Contents = NULL;
      else
      {
	Contents = (ElementT *) malloc(sizeof(ElementT) * Unit * size);
	if (Contents == NULL)
	  throw bad_alloc();
      }

      tSizeChangeNotifier::setSize(size);
    }

    void setUnit(unsigned unit)
    {
      if (unit != Unit)
      {
        Unit = unit;
        setSizeInternal(NumberOf);
      }
    }

    void set(unsigned index, ElementT value)
    {
      if (index >= NumberOf * Unit)
	throw runtime_error("index out of bounds");
      if (Contents == NULL)
	throw runtime_error("Array unallocated");
      Contents[ index ] = value;
    }

    void setSub(unsigned index, unsigned sub_index, ElementT value)
    {
      set(index * Unit + sub_index, value);
    }

    ElementT get(unsigned index)
    {
      if (index >= NumberOf * Unit)
	throw runtime_error("index out of bounds");
      if (Contents == NULL)
	throw runtime_error("Array unallocated");
      return Contents[ index ];
    }

    ElementT getSub(unsigned index, unsigned sub_index)
    {
      return get(index * Unit + sub_index);
    }

    tForeignArray &operator=(tForeignArray const &src)
    {
      if (SlaveTo)
        assert(src.size() == SlaveTo->size());
      else
        setSize(src.size());

      setUnit(src.Unit);

      if (src.Contents)
        memcpy(Contents, src.Contents, sizeof(ElementT) * Unit * src.size());
      else
      {
        free(Contents);
        Contents = NULL;
      }

      return *this;
    }
};




struct tTriangulationParameters : public triangulateio, public boost::noncopyable
{
  public:
    tForeignArray<REAL>		Points; // in/out
    tForeignArray<REAL>		PointAttributes; // in/out
    tForeignArray<int>		PointMarkers; // in/out

    tForeignArray<int>		Triangles; // in/out
    tForeignArray<REAL>		TriangleAttributes; // in/out
    tForeignArray<REAL>		TriangleAreas; // in only
    tForeignArray<int>		Neighbors; // out only

    tForeignArray<int>		Segments; // in/out
    tForeignArray<int>		SegmentMarkers; // in/out
    
    tForeignArray<REAL>		Holes; // in only

    tForeignArray<REAL>		Regions; // in only

    tForeignArray<int>		Edges; // out only
    tForeignArray<int>		EdgeMarkers; // out only
    tForeignArray<REAL>		Normals; // out only

  public:
    tTriangulationParameters()
      : Points("Points", pointlist, numberofpoints, 2),
        PointAttributes("PointAttributes", pointattributelist, numberofpoints, 0, &Points),
	PointMarkers("PointMarkers", pointmarkerlist, numberofpoints, 1, &Points),

	Triangles("Triangles", trianglelist, numberoftriangles, 3),
	TriangleAttributes("TriangleAttributes", triangleattributelist, numberoftriangles, 0, &Triangles),
	TriangleAreas("TriangleAreas", trianglearealist, numberoftriangles, 1, &Triangles),
	Neighbors("Neighbors", neighborlist, numberoftriangles, 3, &Triangles),

	Segments("Segments", segmentlist, numberofsegments, 2),
	SegmentMarkers("SegmentMarkers", segmentmarkerlist, numberofsegments, 1, &Segments),

	Holes("Holes", holelist, numberofholes, 2),

	Regions("Regions", regionlist, numberofregions, 4),

	Edges("Edges", edgelist, numberofedges, 2),
	EdgeMarkers("EdgeMarkers", edgemarkerlist, numberofedges, 1, &Edges),
	Normals("Normals", normlist, numberofedges, 2, &Edges)
    {
      numberofpointattributes = 0;
      numberofcorners = 3;
      numberoftriangleattributes = 0;
    }

    unsigned numberOfPointAttributes() const
    {
      return numberofpointattributes;
    }

    unsigned numberOfTriangleAttributes() const
    {
      return numberoftriangleattributes;
    }

    void setNumberOfPointAttributes(unsigned attrs)
    {
      PointAttributes.setUnit(attrs);
      numberofpointattributes = attrs;
    }

    void setNumberOfTriangleAttributes(unsigned attrs)
    {
      TriangleAttributes.setUnit(attrs);
      numberoftriangleattributes = attrs;
    }

    tTriangulationParameters &operator=(const tTriangulationParameters &src)
    {
      numberofpointattributes = src.numberofpointattributes ;
      numberofcorners = src.numberofcorners;
      numberoftriangleattributes = src.numberoftriangleattributes;

      Points = src.Points;
      PointAttributes = src.PointAttributes;
      PointMarkers = src.PointMarkers;

      Triangles = src.Triangles;
      TriangleAttributes = src.TriangleAttributes;
      TriangleAreas = src.TriangleAreas;
      Neighbors = src.Neighbors;

      Segments = src.Segments;
      SegmentMarkers = src.SegmentMarkers;

      Holes = src.Holes;

      Regions = src.Regions;

      Edges = src.Edges;
      EdgeMarkers = src.EdgeMarkers;
      Normals = src.Normals;

      return *this;
    }
};




tTriangulationParameters *copyTriangulationParameters(const tTriangulationParameters &src)
{
  auto_ptr<tTriangulationParameters> copy(new tTriangulationParameters);
  *copy = src;
  return copy.release();
}




object RefinementFunction;




int triunsuitable(vertex triorg, vertex tridest, vertex triapex, REAL area)
{
  // return 1 if triangle is too large, 0 otherwise
  try
  {
    tVertex org(triorg);
    tVertex dest(tridest);
    tVertex apex(triapex);
    return extract<bool>(RefinementFunction(
	  boost::ref(org), boost::ref(dest), boost::ref(apex), area));
  }
  catch (exception &ex)
  {
    cerr 
      << "*** Oops. Your Python refinement function raised an exception." << endl
      << "*** " << ex.what() << endl
      << "*** Sorry, we can't continue." << endl;
    abort();
  }
  catch (...)
  {
    cerr 
      << "*** Oops. Your Python refinement function raised an exception." << endl
      << "*** Sorry, we can't continue." << endl;
    abort();
  }
}




void triangulateWrapper(char *options, tTriangulationParameters &in, 
    tTriangulationParameters &out,
    tTriangulationParameters &voronoi,
    object refinement_func)
{
  RefinementFunction = refinement_func;
  triangulate(options, &in, &out, &voronoi);
  RefinementFunction = object(); // i.e. None

  out.holelist = NULL;
  out.numberofholes = 0;

  out.regionlist = NULL;
  out.numberofregions = 0;
}




BOOST_PYTHON_MODULE(internals)
{
  def("triangulate", triangulateWrapper);
  class_<tTriangulationParameters, boost::noncopyable>
    ("tTriangulationParameters", init<>())
    .def_readonly("Points", &tTriangulationParameters::Points)
    .def_readonly("PointAttributes", &tTriangulationParameters::PointAttributes)
    .def_readonly("PointMarkers", &tTriangulationParameters::PointMarkers)

    .def_readonly("Triangles", &tTriangulationParameters::Triangles)
    .def_readonly("TriangleAttributes", &tTriangulationParameters::TriangleAttributes)
    .def_readonly("TriangleAreas", &tTriangulationParameters::TriangleAreas)
    .def_readonly("Neighbors", &tTriangulationParameters::Neighbors)

    .def_readonly("Segments", &tTriangulationParameters::Segments)
    .def_readonly("SegmentMarkers", &tTriangulationParameters::SegmentMarkers)

    .def_readonly("Holes", &tTriangulationParameters::Holes)

    .def_readonly("Regions", &tTriangulationParameters::Regions)

    .def_readonly("Edges", &tTriangulationParameters::Edges)
    .def_readonly("EdgeMarkers", &tTriangulationParameters::EdgeMarkers)

    .def_readonly("Normals", &tTriangulationParameters::Normals)

    .add_property("NumberOfPointAttributes", 
        &tTriangulationParameters::numberOfPointAttributes,
        &tTriangulationParameters::setNumberOfPointAttributes)
    .add_property("NumberOfTriangleAttributes", 
        &tTriangulationParameters::numberOfTriangleAttributes,
        &tTriangulationParameters::setNumberOfTriangleAttributes)

    .def("copy", &copyTriangulationParameters,
        return_value_policy<manage_new_object>())
    .enable_pickling(true)
    ;
  
  class_<tForeignArray<REAL>, boost::noncopyable>
    ("tRealArray", no_init)
    .def("size", &tForeignArray<REAL>::size)
    .def("setSize", &tForeignArray<REAL>::setSize)
    .def("setup", &tForeignArray<REAL>::setup)
    .def("unit", &tForeignArray<REAL>::unit)
    .def("set", &tForeignArray<REAL>::set)
    .def("setSub", &tForeignArray<REAL>::setSub)
    .def("get", &tForeignArray<REAL>::get)
    .def("getSub", &tForeignArray<REAL>::getSub)
    .def("deallocate", &tForeignArray<REAL>::deallocate)
    ;
	
  class_<tForeignArray<int>, boost::noncopyable>
    ("tIntArray", no_init)
    .def("size", &tForeignArray<int>::size)
    .def("setSize", &tForeignArray<int>::setSize)
    .def("setup", &tForeignArray<int>::setup)
    .def("unit", &tForeignArray<int>::unit)
    .def("set", &tForeignArray<int>::set)
    .def("setSub", &tForeignArray<int>::setSub)
    .def("get", &tForeignArray<int>::get)
    .def("getSub", &tForeignArray<int>::getSub)
    .def("deallocate", &tForeignArray<int>::deallocate)
    ;

  class_<tVertex, bases<>, tVertex, boost::noncopyable>("tVertex", no_init)
    .def("x", &tVertex::x)
    .def("y", &tVertex::y)
    ;
}
